% Two-dimensional, incomplete data FDE test, solved by TT-MINRES method

%%%%%%%%%%% Setup part begin %%%%%%%%%%


alpha = 0.5;    % Temporal derivative order
beta1 = 1.5; beta2 = 1.5;  % Space derivative orders

L = 1;      % length of spatial interval
d = 4;      % grid level
nx = 2^d+2; % Number of spatial steps of discretization
ny = 2^d+2;

nt = 2^(2*d);   % Number of steps in time


tol = 1e-6;     % tensor approximation for the solution

eps = 1e-14;    % tensor approximation for matrices.
                % It should be as small as possible, but
                % it cannot be the machine eps due to numerical SVD noise

gamma=1e-6;     % Regularization parameter 
                % !!!! we will divide over it, do not set zero !!!

% Rectangular observation domain, defined by left and right grid points,
% i.e. [ax_obs,bx_obs] x [ay_obs,by_obs]
% The values should be given, assuming the boundary points excluded!

ax_obs = 2^(d-1)-2^(d-3);   % left bound for observation domain
bx_obs = 2^(d-1)+2^(d-3)-1; % right bound for observation domain
ay_obs = 2^(d-1)-2^(d-3);   % lower bound for obs domain
by_obs = 2^(d-1)+2^(d-3)-1; % upper bound for obs domain

% Rectangular control domain, defined by left and right grid points,
% i.e. [ax_con,bx_con] x [ay_con,by_con]
% The values should be given, assuming the boundary points excluded!

ax_con = 2^(d-1)-2^(d-3);   % left bound for observation domain
bx_con = 2^(d-1)+2^(d-3)-1; % right bound for observation domain
ay_con = 2^(d-1)-2^(d-3);   % lower bound for obs domain
by_con = 2^(d-1)+2^(d-3)-1; % upper bound for obs domain

%%%%%%%%%%% Setup part end %%%%%%%%%%

% Auto-initialise TT-Toolbox
if (exist('tt_tensor', 'file')==0)    
    % Check whether the TT-Toolbox directory is already here
    if (exist('TT-Toolbox-small', 'dir')~=7)
        fprintf('Downloading and unpacking TT-Toolbox...\n');
        [f,stat] = urlwrite('https://codeload.github.com/oseledets/TT-Toolbox/zip/small', 'TT-Toolbox-small.zip');
        if (stat==0)
            error('Sorry, automatic downloading failed. You need to download http://github.com/oseledets/TT-Toolbox/archive/small.zip and unzip it manually.');
        end;
        unzip('TT-Toolbox-small.zip');
        delete('TT-Toolbox-small.zip');
    end;
    fprintf('Adding TT-Toolbox to the path...\n');
    cd TT-Toolbox-small
    setup
    cd ..
    fprintf('TT-Toolbox is ready\n');
end;

hx = L/(nx-1);  % spatial mesh steps
hy = L/(ny-1);

tau = hx*hy;    % time step

nu = hx*hy*tau*gamma; % parameter only needed for the approximation of the singular block, only relevant for MinRes

% Quantized sizes
mx = factor(nx-2); % space-x size
my = factor(ny-2); % space-y size
mt = factor(nt);   % temporal size
mx_c = factor(bx_con-ax_con+1); % control dimension x
my_c = factor(by_con-ay_con+1); % control dimension y



% Coefficient vectors by Grunwald-Letnikov formulae
v_alpha = cumprod([1, 1 - ((alpha+1) ./ (1:nt))]);
v_beta1 =  cumprod([1, 1 - ((beta1+1) ./ (1:nx))]);
v_beta2 = cumprod([1, 1 - ((beta2+1) ./ (1:ny))]);

% Temporal Toeplitz matrix in the QTT format
% Compress v_alpha i.e. the first column
tt_v_alpha = tt_tensor(v_alpha(1:end-1), eps, mt);
% The tt_qtoepl routine puts elements (n+1)--(2n) to the column. 
% Shift v_alpha accordingly
tt_v_alpha = tkron(tt_v_alpha, tt_tensor([0;1]));
% Create the Toeplitz matrix
tt_Talpha = (tau^(-alpha))*tt_qtoepl(tt_v_alpha, log2(nt));
tt_Talpha = round(tt_Talpha, eps);

% Spatial matrices in the QTT format. 
% Boundary elements are already excluded.
tt_v_beta1 = tt_tensor(v_beta1(2:end-2), eps, mx);
% Shift to the range (n+1)--(2n)
tt_v_beta1 = tkron(tt_v_beta1, tt_tensor([0;1]));
% Add one element above the diagonal
tt_v_beta1 = tt_v_beta1 + v_beta1(1)*tt_unit(2,d+1,2^d);
tt_v_beta1 = round(tt_v_beta1, eps);
% Create the Toeplitz matrix
tt_C_beta1 = (hx^(-beta1))*tt_qtoepl(tt_v_beta1, d);
tt_L_beta1 = (tt_C_beta1+tt_C_beta1')*0.5;
tt_L_beta1 = round(tt_L_beta1, eps);

% Repeat the same steps for other coordinates
tt_v_beta2 = tt_tensor(v_beta2(2:end-2), eps, my);
tt_v_beta2 = tkron(tt_v_beta2, tt_tensor([0;1]));
tt_v_beta2 = tt_v_beta2 + v_beta2(1)*tt_unit(2,d+1,2^d);
tt_v_beta2 = round(tt_v_beta2, eps);
tt_C_beta2 = (hy^(-beta2))*tt_qtoepl(tt_v_beta2, d);
tt_L_beta2 = (tt_C_beta2+tt_C_beta2')*0.5;
tt_L_beta2 = round(tt_L_beta2, eps);


% A tensor-structured version of meshgrid
tt_X = (tt_x(2,d)+1)*hx;
tt_X = mtkron(tt_X, tt_ones(2,d));
tt_Y = (tt_x(2,d)+1)*hy;
tt_Y = mtkron(tt_ones(2,d), tt_Y);

% The forcing vector is zero
tt_F = tt_zeros(2, d*2);

% The target vector is constructed as a function of X,Y,Z via the cross
% interpolation.
tt_YD = amen_cross({tt_X,tt_Y}, @(X)(10*cos(10*X(:,1)).*sin(X(:,2).*X(:,1))), eps);



% Create an observation operator
% Prepare the vectors of ones between ax and bx and zeros otherwise
c1 = zeros(nx-2, 1);
c1(ax_obs:bx_obs) = 1;
% And compress them to tt_tensors.
c1 = tt_tensor(c1, eps, mx);

c2 = zeros(ny-2, 1);
c2(ay_obs:by_obs) = 1;
c2 = tt_tensor(c2, eps, my);

% Put them onto a diagonal
tt_C1 = diag(c1);
tt_C2 = diag(c2);

% Complement to c1 -- needed in the preconditioning of Minres
c1app = zeros(nx-2, 1);
c1app(1:ax_obs-1) = 1;
c1app(bx_obs+1:end) = 1;
c1app = tt_tensor(c1app, eps, mx);

c2app = zeros(ny-2, 1);
c2app(1:ay_obs-1) = 1;
c2app(by_obs+1:end) = 1;
c2app = tt_tensor(c2app, eps, my);

tt_C1app = diag(c1app);
tt_C2app = diag(c2app);


% Create a control operator
% These matrices are rectangular, prepare them as such, putting an identity
% matrix to the block corresponding to [ax,bx], etc.
p1=sparse(nx-2,bx_con-ax_con+1);
p1(ax_con:bx_con,:) = speye(bx_con-ax_con+1);
p2=sparse(ny-2,by_con-ay_con+1);
p2(ay_con:by_con,:) = speye(by_con-ay_con+1);
% Now they should be unfortunately compressed from the full format, since the
% system and control sizes can have different factors
tt_p1 = tt_matrix(full(p1), eps, mx, mx_c);
tt_p2 = tt_matrix(full(p2), eps, my, my_c);


% Spatial FDE matrix = L1 x I x I + I x L2 x I + I x I x L3
Lspace = mtkron(tt_L_beta1, tt_eye(my))+...
    mtkron(tt_eye(mx), tt_L_beta2);

% Temporal FDE matrix = (IxIxI) x T
Ltime = tkron(tt_eye([mx,my]), tt_Talpha);
% Spatial matrix extended by I in the time variable
A = tkron(Lspace,tt_eye(mt));
% Finally, the whole system matrix is as follows
A=(hx*hy*tau)*round(Ltime-A,eps);

% Observation operator -- the kron of Ci and the identity in time.
M1 = mtkron(tt_C1, tt_C2, tt_eye(mt));
M1 = (hx*hy*tau)*M1;

% A surrogate for M1, needed for the preconditioner
M1app = hx*hy*tau*mtkron(tt_C1, tt_C2, tt_eye(mt))+...
    mtkron(nu*tt_C1, tt_C2app, tt_eye(mt))+...
    mtkron(tt_C1app, nu*tt_C2, tt_eye(mt))+...
    mtkron(sqrt(nu)*tt_C1app, sqrt(nu)*tt_C2app, tt_eye(mt));    

% A surrogate for M1 inversed, needed for the preconditioner
invM1 = 1/(hx*hy*tau)*mtkron(tt_C1, tt_C2, tt_eye(mt))+...
    mtkron((1/nu)*tt_C1, tt_C2app, tt_eye(mt))+...
    mtkron(tt_C1app, 1/(nu)*tt_C2, tt_eye(mt))+...
    mtkron(1/sqrt(nu)*tt_C1app, 1/sqrt(nu)*tt_C2app, tt_eye(mt));                      


% Rectangular control--system "interaction" operator
M3 = mtkron(tt_p1, tt_p2, tt_eye(mt));
M3 = (hx*hy*tau)*M3;

% A surrogate for M3, needed for the preconditioner
M3app =((hx*hy*tau)/sqrt(gamma)*mtkron(tt_C1*(tt_p1*tt_p1'), tt_C2*(tt_p2*tt_p2'), tt_eye(mt))+...
                         mtkron(sqrt(hx*hy*tau/gamma)*sqrt((nu))*tt_C1*(tt_p1*tt_p1'), tt_C2app*(tt_p2*tt_p2'), tt_eye(mt))+...
                         mtkron(tt_C1app*(tt_p1*tt_p1'), sqrt(hx*hy*tau/gamma)*sqrt((nu))*tt_C2*(tt_p2*tt_p2'), tt_eye(mt))+...
                         mtkron(sqrt(sqrt(hx*hy*tau/gamma))*sqrt(sqrt(nu))*tt_C1app*(tt_p1*tt_p1'), sqrt(sqrt(hx*hy*tau/gamma))*sqrt(sqrt(nu))*tt_C2app*(tt_p2*tt_p2'), tt_eye(mt)));
                     

% Control mass matrix
M2 = tt_eye([mx_c, my_c, mt])*(hx*hy*tau*gamma);
% Inversed control matrix
invM2 = tt_eye([mx_c, my_c, mt])/(hx*hy*tau*gamma);

% Right hand side, first block
pre_f1 = tkron(tt_YD, tt_ones(mt));
f1 = M1*pre_f1;
% The second block is always zero

% Third block
pre_f3 = tkron(tt_F, tt_ones(mt));
f3 = hx*hy*tau*pre_f3;


%%% The MinRes solver

fprintf('\n======== tt-minres\n');
tic;
% Create factors of a preconditioner, i.e. S=S1*iM1*S2
S1 = (A+M3app);
S1 = round(S1, eps);
S2 = (A'+M3app');
S2 = round(S2, eps);

% RHS in the blocked form
Pb = cell(1,3);
Pb{1} = f1;
Pb{2} = mtkron(tt_zeros(mx_c,1),tt_zeros(my_c,1),tt_zeros(mt,1));
Pb{3} = f3;
ttimes(1,3)=toc;

% Finally, solve
tic;
sol_mr = tt_minres(@(x,tol)Afun(x, tol, A, M1, M2, M3), ...
    Pb, tol/3, 'P', @(x,tol)Pfun(x, tol, M1app, invM1, invM2, S1, S2), ...
    'relaxation', 0.5, 'tol_exit', tol, 'max_iters', 100);
ttimes(2,3)=toc;

% Now, y = sol_mr{1}, u = sol_mr{2}, p = sol_mr{3}.

% We need to reshape u to measure residuals
sol_mr{2} = tt_reshape(sol_mr{2}, M3.m);
resid_row1 = norm(M1*sol_mr{1}+A'*sol_mr{3} - f1)/norm(f1)
M2r = M2;
M2r = tt_reshape(M2r, M3.m*[1,1]);
resid_row2 = norm(M2r*sol_mr{2}+M3'*sol_mr{3})/norm(M3'*sol_mr{3})
resid_row3 = norm(A*sol_mr{1} + M3*sol_mr{2} - f3)/max(norm(f3), norm(M3*sol_mr{2}))

obs_error = norm(f1-M1*sol_mr{1})/norm(f1)


fprintf('\n======== tt-minres completed, timings: [%g,%g]\n', ttimes(1,3), ttimes(2,3));


% For small grids, we can visualise this
if (d<=8)
    % Create the grid
    [X2,Y2] = meshgrid(hx:hx:L-hx,hy:hy:L-hy);
    % A matrix for extracting the last time snapshot
    P = mtkron(tt_eye(mx), tt_eye(my), tt_matrix(tt_unit(mt, numel(mt), nt), 1, mt));
    
    % State in the full format
    y_full = P*(M1/(hx*hy*tau))*sol_mr{1};
    y_full = full(y_full, [nx-2,ny-2]);
    figure(1);
    mesh(X2,Y2,y_full);
    title('y');
    
    % Target in the full format
    yd_full = P*f1/(hx*hy*tau);
    yd_full = full(yd_full, [nx-2,ny-2]);
    figure(2)
    mesh(X2,Y2,yd_full);
    title('yd');
    
    % Control
    u_full = P*(M3/(hx*hy*tau))*sol_mr{2};
    u_full = full(u_full, [nx-2,ny-2]);
    figure(3)
    mesh(X2,Y2,u_full);
    title('u');
    
    % Lagrange multipliers
    p_full = P*sol_mr{3};
    p_full = full(p_full, [nx-2,ny-2]);
    figure(4)
    mesh(X2,Y2,p_full);
    title('p');
end;

