function [x,td] = tt_minres(A, b, tol, varargin)
%Symmetric TT-GMRES (minres) method with a separated block storage of the solution.
%   [x,td] = tt_minres(A, b, tol, varargin)
%
% Tries to solve a linear system Ax=b, where x and b are cell arrays of
% size 1 x M, each b{m} is a tt_tensor, such that the total vector would
% write as [b{1}; ... ; b{M}].
%
% A should be given as a function handle @(x,tol)A(x), which takes and return
% cell arrays of tt_tensors of the form given above, and computes the
% Matrix-Vector product with the approximation tolerance tol. 
%
% The same holds true for the (optional) preconditioner P in varargin
%
% A, P and b can be passed as regular tt_matrix and tt_tensor, resp. However,
% the result will still be returned in the form cell(1,1).
%
% varargin may contain a sequence of tuning parameters of the form 
% 'parameter1_name', parameter1_value, 'parameter2_name', parameter2_value
% and so on. Available parameters and [default values] are the following.
%   'relaxation' [0]: Inexact Krylov relaxation power. The accuracy for 
%                       Krylov vectors is set as tol/(|r|/|b|)^relaxation, 
%                       where r is the current residual
%   'max_iters' [100]: Number of iterations
%   'x0' [all-zeros vector]: Initial guess
%   'verb' [2]:  Verbosity. 0 - Silent, 1 - outer iteration info, 2 - inner
%                iteration info, 3 - iterations + return all solutions in td{2}
%   'tol_exit' [tol]:  Stopping tolerance: exit if |r|/|b|<tol_exit
%   'p' []: Preconditioning function/matrix
%            
%
%   Debug output td is currently syncronised with the amen and dmrg:
%       td{1} - cumulative CPU time,
%       td{2} - current solutions x,
%       td{3} - local residuals
%
%
% Development: Sergey Dolgov ( sergey.v.dolgov@gmail.com ),
% Max Planck Institute for Mathematics in the Sciences, Leipzig.
%
%TT-Toolbox 2.2, 2009-2012
%
%This is TT Toolbox, written by Ivan Oseledets et al.
%Institute of Numerical Mathematics, Moscow, Russia
%webpage: http://spring.inm.ras.ru/osel
%
%For all questions, bugs and suggestions please mail
%ivan.oseledets@gmail.com
%---------------------------


% Inexact Krylov relaxation power. The accuracy for Krylov vectors is set
% as tol/(|r|/|b|)^relaxation, where r is the current residual
relaxation = 0;
% Number of outer iterations
max_iters = 100;
% Initial guess
x = [];
% Verbosity
verb = 2;
% Stopping tolerance
tol_exit = tol;
% Prec
P = [];

for i=1:2:length(varargin)-1
    switch lower(varargin{i})
        case 'relaxation'
            relaxation = varargin{i+1};
        case 'max_iters'
            max_iters = varargin{i+1};
        case 'x0'
            x = varargin{i+1};
        case 'verb'
            verb = varargin{i+1};
        case 'tol_exit'
            tol_exit = varargin{i+1};       
        case 'p'
            P = varargin{i+1};
        otherwise
            error('Unknown tuning parameter "%s"', varargin{i});
    end;
end;

% Block size
if (isa(b, 'tt_tensor'))
    M = 1;
    r = {b};
else
    M = numel(b);
    r = b;
end;

% If we are given a single tt_matrix, just use exact MV+round
if (isa(A, 'tt_matrix'))
    A = @(x,tol){round(A*x{1}, tol)};
end;
if (~isempty(P))&&(isa(P, 'tt_matrix'))
    P = @(x,tol){round(P*x{1}, tol)};
end;

% Debug info
if (nargout>1)
    td = cell(3,1);
    td{1} = zeros(1, max_iters);
    td{2} = cell(1, max_iters);
    td{3} = zeros(1, max_iters);
end;

% Check for the initial guess
if (isempty(x))
    x = cell(1,M);
    for m=1:M
        x{m} = tt_zeros(r{m}.n);
    end;
else
    if (isa(x,'tt_tensor'))
        x = {x};
    end;
end;

t_minres_start = tic;

% Initialize.
% numbers
c0 = 1; c1 = 1; s0 = 0; s1 = 0; gamma0 = 1;
tol_kryl = tol;
% vectors
% initial residual
v2 = r;
if (cellnorm(x)>0)
    v1 = A(x, tol);
    for m=1:M
        v2{m} = round(v2{m}-v1{m}, tol);
    end;
end;
z1 = v2;
if (~isempty(P))
    z1 = P(v2, tol);
end;
gamma1 = sqrt(abs(celldot(v2, z1)));
eta = gamma1; % It will measure the local residuals
eta0 = eta;

% Other Krylov vectors. Only 3-term
v1 = cell(1,M);
for m=1:M
    v1{m} = tt_zeros(r{m}.n);
end;
w1 = v1; w2 = v1;

for it=1:max_iters
    max_x_rank = 0;
    max_v_rank = 0;

    % normalise z
    for m=1:M
        z1{m} = z1{m}/gamma1;
    end;
    % Apply MatVec
    v3 = A(z1, tol_kryl);
    % Subtract 3-term orthogonalisation
    delta = celldot(z1, v3);
    for m=1:M
        v3{m} = v3{m} - (delta/gamma1)*v2{m} - (gamma1/gamma0)*v1{m};
        v3{m} = round(v3{m}, tol_kryl);
        max_v_rank = max(max_v_rank, max(v3{m}.r));
    end;
    
    % Apply prec
    z2 = v3;
    if (~isempty(P))
        z2 = P(v3, tol_kryl);
    end;
    
    % Solve the least-squares tridiag. system by hands
    gamma2 = sqrt(abs(celldot(v3, z2)));
    alpha0 = c1*delta-c0*s1*gamma1;
    alpha1 = sqrt(alpha0^2+gamma2^2);
    alpha2 = s1*delta + c0*c1*gamma1;
    alpha3 = s0*gamma1;
    c0 = c1; c1 = alpha0/alpha1; 
    s0 = s1; s1 = gamma2/alpha1;
    
    % Orthogonalise the preconditioned Krylov vector
    w3 = z1;
    for m=1:M
        w3{m} = w3{m} - alpha3*w1{m} - alpha2*w2{m};
        w3{m} = round(w3{m}, tol_kryl);   
        w3{m} = w3{m}/alpha1;
        max_v_rank = max(max_v_rank, max(w3{m}.r));
    end;
   
    % Correct the solution
    for m=1:M
        x{m} = x{m} + (c1*eta)*w3{m};
        x{m} = round(x{m}, tol);
        max_x_rank = max(max_x_rank, max(x{m}.r));
    end;

    % Update the residual
    eta = -s1*eta;    
    % Relax the Krylov tolerance
    tol_kryl = min(tol/(abs(eta/eta0)^relaxation), 0.9);
    
    % Push vectors to the history
    v1 = v2; v2 = v3; z1 = z2; w1 = w2; w2 = w3;
    % And numbers
    gamma0 = gamma1; gamma1 = gamma2;
    
    % Various reports and checking    
    if (nargout>1)
        % report the residual
        td{3}(it)=abs(eta/eta0);
        % Report time
        td{1}(it) = toc(t_minres_start);
        % Report the current solution, if required
        if (verb>2)
            td{2}{it} = x;
        end;
    end;

    if (verb>0)
        fprintf('==tt_minres== iter=%d, |r| = %g, rank_v=%d, rank_x=%d, time=%g  \n', it, abs(eta/eta0), max_v_rank, max_x_rank, toc(t_minres_start));
    end;    
    
    if (abs(eta/eta0)<tol_exit); break; end;
end;

if (nargout>1)
    td{1}=td{1}(1:it);
    td{2}=td{2}(1:it);
    td{3}=td{3}(1:it);
end;

end

% Compute the second norm of the cell of tt_tensors
function nrm = cellnorm(x)
nrm = 0;
M = numel(x);
for m=1:M
    nrm = nrm+norm(x{m})^2;
end;
nrm = sqrt(nrm);
end

% Compute the dot product of cells of tt_tensors
function sc = celldot(x,y)
sc = 0;
M = numel(x);
for m=1:M
    sc = sc+dot(x{m},y{m});
end;
end